/**
 *  Dining philosophers example. The program uses Java threads and locks provided by the synchronized keyword.
 *  Output on console -- text mode application. The program terminates when all philosophers have already eaten
 *  and thought 5 times. You can redefine the Philosopher.NO_OF_TURNS constant or rewrite the Philosopher.run()
 *  method (replace for(...) with while(true)) to change this behavior.
 *
 *  @author <a href="http://home.zcu.cz/~jkacer">Jarda Kacer</a>, Czech OS/2 User Group; email: <a href="mailto:jkacer@kiv.zcu.cz">jkacer@kiv.zcu.cz</a>
 *
 */
public class DiningPhilosophers
{
 private int            noOfLivingPhilosophers;
 private int            timeToThink;
 private int            timeToEat;

 private Fork[]         forks;
 private Philosopher[]  philosophers;


 /**
  *  Runs the program.
  *
  */
 public static void main(String[] args)
 {
  DiningPhilosophers    dp;

  dp = new DiningPhilosophers();
  dp.doIt(args);
 } // main


 /**
  *  If command line parameters are accepted, creates philosophers, lets them run and waits until they all finish.
  *
  */
 private void doIt(String[] args)
 {
  noOfLivingPhilosophers = 0;
  hello();
  if (args.length != 2)
   usage();
  else
  {
   if (setTimes(args))
   {
    printExplanation();
    createThreads();
    waitForZeroPhilosophers();
    System.out.println("Main thread: Good-bye!");
   } // if setTimes
  } // else args
 } // doIt


 /**
  *  Prints out a hello message.
  *
  */
 private void hello()
 {
  System.out.println("Dining Philosophers Example -- Java version");
  System.out.println("Author: Jarda Kacer, jkacer@kiv.zcu.cz, Czech OS/2 User Group\n");
 } // hello


 /**
  *  Processes command line parameters.
  *
  */
 private boolean setTimes(String[] args)
 {
  boolean error = false;

  try
  {
   timeToThink = 1000 * Integer.valueOf(args[0]).intValue();
   timeToEat   = 1000 * Integer.valueOf(args[1]).intValue();
   System.out.println("Using " + timeToThink + " ms as maximum thinking time and " + timeToEat + " ms as maximum eating time.");
  } // try
  catch (NumberFormatException e)
  {
   System.out.println("Invalid parameters specified.");
   error = true;
  } // catch

  return (!error);
 } // setTimes


 /**
  *  Prints out an explanation of symbols used during program output.
  *
  */
 private void printExplanation()
 {
  System.out.println();
  System.out.println(Philosopher.STATES[Philosopher.STATE_THINKING]  + " ... Thinking");
  System.out.println(Philosopher.STATES[Philosopher.STATE_WAITING]   + " ... Waiting for a fork");
  System.out.println(Philosopher.STATES[Philosopher.STATE_EATING]    + " ... Eating");
  System.out.println(Philosopher.STATES[Philosopher.STATE_RELEASING] + " ... Releasing a fork");
  System.out.println(Philosopher.STATES[Philosopher.STATE_DEAD]      + " ... Dead (without any fork)");
  System.out.println();
  System.out.println("| X | ... The philosopher has no fork.");
  System.out.println(" /X | ... The philosopher has his right fork.");
  System.out.println("| X\\  ... The philosopher has his left fork.");
  System.out.println(" /X\\  ... The philosopher has both forks.");
  System.out.println();
 } // printExplanation


 /**
  *  Creates all philosophers.
  *
  */
 private void createThreads()
 {
  int   i;
  Fork  left;
  Fork  right;

  forks = new Fork[Philosopher.NO_OF_PHILOSOPHERS];
  for (i = 0; i < Philosopher.NO_OF_PHILOSOPHERS; i++)
   forks[i] = new Fork();

  philosophers = new Philosopher[Philosopher.NO_OF_PHILOSOPHERS];
  for (i = 0; i < Philosopher.NO_OF_PHILOSOPHERS; i++)
  {
   if (i == 0)
    right = forks[Philosopher.NO_OF_PHILOSOPHERS - 1];
   else
    right = forks[i - 1];
   left = forks[i];

   philosophers[i] = new Philosopher(this, i, left, right, timeToThink, timeToEat);
  } // for

  for (i = 0; i < Philosopher.NO_OF_PHILOSOPHERS; i++)
   philosophers[i].start();

  noOfLivingPhilosophers = Philosopher.NO_OF_PHILOSOPHERS;
 } // createThreads


 /**
  *  Prints out a message explaining program's usage.
  *
  */
 private void usage()
 {
  System.out.println("Usage: java " + getClass().getName() + " MAX_THINKING_TIME MAX_EATING_TIME");
  System.out.println("MAX_THINKING_TIME ... Maximum time (in seconds) the philosophers will think.");
  System.out.println("MAX_EATING_TIME ..... Maximum time (in seconds) the philosophers will eat.\n");
  System.out.println("The actual time used for thinking/eating is a random number");
  System.out.println("between 0 and MAX_?_TIME which is generated every time a philosopher");
  System.out.println("attempts to think or eat.\n");
 } // usage


 /**
  *  Prints out a line describing the current state of forks and philosophers.
  *
  */
 public synchronized void printForks()
 {
  for (int i = 0; i < Philosopher.NO_OF_PHILOSOPHERS; i++)
  {
   if (philosophers[i].getRightFork().getOwner() == philosophers[i])
    System.out.print(" /");
   else
    System.out.print("  ");

   System.out.print(Philosopher.STATES[philosophers[i].getState()]);

   if (philosophers[i].getLeftFork().getOwner() == philosophers[i])
    System.out.print("\\ ");
   else
    System.out.print("  ");

   if (forks[i].getOwner() == null)
    System.out.print("|");
   else
    System.out.print(" ");
  } // for i

  System.out.println();
 } // printForks


 /**
  *  Decrements the count of philosophers present in the program.
  *
  */
 public synchronized void decrementCountOfPhilosophers()
 {
  noOfLivingPhilosophers--;
  if (noOfLivingPhilosophers == 0)
   notify();
 } // decrementCountOfPhilosophers


 /**
  *  Blocks until all philosophers finish their work.
  *
  */
 private synchronized void waitForZeroPhilosophers()
 {
  while (noOfLivingPhilosophers > 0)
  {
   try
   {
    wait();
   } // try
   catch (InterruptedException e)
   {
    System.err.println("The main thread interrupted while waiting for the philosophers to finish!");
   } // catch
  } // while
 } // waitForZeroPhilosophers
} // class DiningPhilosophers

