/**
 *  A philosopher spends its life by thinking and eating. To be able to eat, he must get both forks, one on his
 *  left side and one on his right side. After finishing eating, he releases both forks. He dies after having
 *  completed 5 life cycles. In a real example, the loop would be infinite.
 *
 *  @author <a href="http://home.zcu.cz/~jkacer">Jarda Kacer</a>, Czech OS/2 User Group; email: <a href="mailto:jkacer@kiv.zcu.cz">jkacer@kiv.zcu.cz</a>
 *
 */
public class Philosopher extends Thread
{
 public static final int NO_OF_PHILOSOPHERS     = 5;

 public static final int NO_OF_STATES           = 5;
 public static final int NO_OF_TURNS            = 5;

 public static final int STATE_THINKING         = 0;
 public static final int STATE_WAITING          = 1;
 public static final int STATE_EATING           = 2;
 public static final int STATE_RELEASING        = 3;
 public static final int STATE_DEAD             = 4;

 public static final char[] STATES = {'T', 'W', 'E', 'R', 'D'};

 private int                    id;
 private int                    state;

 private Fork                   leftFork;
 private Fork                   rightFork;

 private Fork                   firstFork;
 private Fork                   secondFork;

 private int                    timeToThink;
 private int                    timeToEat;

 private DiningPhilosophers     parent;


 /**
  *  Creates a new philosopher, with a unique ID.
  *
  */
 public Philosopher(DiningPhilosophers parent, int id, Fork leftFork, Fork rightFork, int timeToThink, int timeToEat)
 {
  super("Philosopher No. " + id);
  this.parent = parent;
  this.id = id;
  this.state = STATE_DEAD;

  this.leftFork = leftFork;
  this.rightFork = rightFork;

  this.timeToThink = timeToThink;
  this.timeToEat = timeToEat;

  if ((id % 2) == 0)
  {
   firstFork = leftFork;
   secondFork = rightFork;
  }
  else
  {
   firstFork = rightFork;
   secondFork = leftFork;
  }
 } // constructor


 /**
  *  Returns the philosopher's state.
  *
  */
 public int getState()
 {
  return state;
 } // getState


 /**
  *  Returns the philosopher's left fork.
  *
  */
 public Fork getLeftFork()
 {
  return leftFork;
 } // getLeftFork


 /**
  *  Returns the philosopher's right fork.
  *
  */
 public Fork getRightFork()
 {
  return rightFork;
 } // getRightFork


 /**
  *  Suspends the philosopher for a random number of seconds -- simulation of thinking.
  *
  */
 private void think()
 {
  try
  {
   sleep((int) (Math.random() * (double) timeToThink));
  } // try
  catch (InterruptedException e)
  {
   System.err.println("A philosopher interrupted when thinking!");
  } // catch
 } // think


 /**
  *  Suspends the philosopher for a random number of seconds -- simulation of eaiting.
  *
  */
 private void eat()
 {
  try
  {
   sleep((int) (Math.random() * (double) timeToEat));
  } // try
  catch (InterruptedException e)
  {
   System.err.println("A philosopher interrupted when eating!");
  } // catch
 } // eat


 /**
  *  Periodically thinks and eats. To be able to eat, the philosopher must get both forks. After finishing
  *  eating, he releases the forks. For simplicity, the number of loops is limited to 5.
  *
  */
 public void run()
 {
  for (int i = 0; i < NO_OF_TURNS; i++)
  {
   // Thinking...
   state = STATE_THINKING;
   parent.printForks();
   think();

   // Trying to get both forks...
   state = STATE_WAITING;
   parent.printForks();

   firstFork.acquireFork(this);
   parent.printForks();

   secondFork.acquireFork(this);
   parent.printForks();

   // Eating
   state = STATE_EATING;
   parent.printForks();
   eat();

   // Releasing the forks...
   state = STATE_RELEASING;
   parent.printForks();

   secondFork.releaseFork(this);
   parent.printForks();

   firstFork.releaseFork(this);
   parent.printForks();
  } // for i

  state = STATE_DEAD;
  parent.printForks();
  parent.decrementCountOfPhilosophers();
 } // run
} // class Philosopher

