/*
Dining philosophers example. The program uses native OS/2 threads and native OS/2 semaphores
accessed via OS/2 API functions. Output on console -- text mode application. The program
terminates when all philosophers have already eaten and thought 5 times. You can redefine
the NO_OF_TURNS constant or rewrite the fnPhilosopherThread function (replace for(...) with while(true))
to change this behavior.

Author: Jarda Kacer, Czech OS/2 User Group
        mailto:jkacer@kiv.zcu.cz
        http://home.zcu.cz/~jkacer
*/


/*------------------------------------------------------------------------------------*/
/* Defined constants */


#define INCL_DOSERRORS
#define INCL_DOSPROCESS
#define INCL_DOSSEMAPHORES

#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <string.h>

#define THREAD_SUSPEND          1L

#define NO_OF_PHILOSOPHERS      5
#define NOBODY                  -1

#define NO_OF_STATES            5

#define STATE_THINKING          0
#define STATE_WAITING           1
#define STATE_EATING            2
#define STATE_RELEASING         3
#define STATE_DEAD              4

#define NO_OF_TURNS             5



/*------------------------------------------------------------------------------------*/
/* Types */


typedef struct _FORK
        {
         INT    iOwner;
         HMTX   hmtxSemaphore;
        }
        FORK;

typedef FORK FORKS[NO_OF_PHILOSOPHERS];

typedef struct _PHILOSOPHER
        {
         INT    iID;
         INT    iState;

         INT    iLeftFork;
         INT    iRightFork;

         INT    iFirstFork;
         INT    iSecondFork;
        }
        PHILOSOPHER;

typedef PHILOSOPHER PHILOSOPHERS[NO_OF_PHILOSOPHERS];

typedef struct _PHILOSOPHER_DATA
        {
         FORKS          *pForks;
         PHILOSOPHERS   *pPhilosophers;
         UINT           uiMyNumber;
        }
        PHILOSOPHER_DATA;



/*------------------------------------------------------------------------------------*/
/* Global data */


HMTX    hmtxPrintMutex;
HMTX    hmtxThreadCountMutex;
INT     iNumberOfAlivePhilosophers;
HEV     hevBlockMainThreadSemaphore;
UCHAR   aucStates[NO_OF_STATES] = {'T', 'W', 'E', 'R', 'D'};
UINT    uiTimeToThink;
UINT    uiTimeToEat;
CHAR    *pszProjectName = "dining_philosophers_c_native_os2.exe";



/*------------------------------------------------------------------------------------*/
/* Functions */


/*------------------------------------------------------------------------------------
print_1_philosopher
------------------------------------------------------------------------------------*/
VOID print_1_philosopher(FORKS forks, PHILOSOPHER philosopher)
{
 /* Right fork */
 if (forks[philosopher.iRightFork].iOwner == philosopher.iID)
  printf(" /");
 else
  printf("  ");

 printf("%c", aucStates[philosopher.iState]);

 /* Left fork */
 if (forks[philosopher.iLeftFork].iOwner == philosopher.iID)
  printf("\\ ");
 else
  printf("  ");
} /* print_1_philosopher */



/*------------------------------------------------------------------------------------
print_forks
------------------------------------------------------------------------------------*/
VOID print_forks(FORKS forks, PHILOSOPHERS philosophers)
{
 APIRET         arReturn;
 INT            i;


 arReturn = DosRequestMutexSem(hmtxPrintMutex, SEM_INDEFINITE_WAIT);
 if (arReturn != 0)
  fprintf(stderr, "\nCannot get a mutex semaphore, return code = %d\n", arReturn);
 else
 {
  for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
  {
   print_1_philosopher(forks, philosophers[i]);
   if (forks[i].iOwner == NOBODY)
    printf("|");
   else
    printf(" ");
  }
  printf("\n");
  arReturn = DosReleaseMutexSem(hmtxPrintMutex);
  if (arReturn != 0)
   fprintf(stderr, "\nCannot release a mutex semaphore, return code = %d\n", arReturn);
 } /* OK */
} /* print_forks */



/*------------------------------------------------------------------------------------
init_data
------------------------------------------------------------------------------------*/
BOOL init_data(FORKS forks, PHILOSOPHERS philosophers)
{
 INT            i;
 APIRET         arReturn;
 BOOL           bError = FALSE;
 ULONG          ulAttributes;
 BOOL32         b32InitState;

 ulAttributes = 0; /* Opposite to DC_SEM_SHARED */
 b32InitState = FALSE; /* Unowned */
 printf("Creating semaphores... ");

 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  forks[i].iOwner = NOBODY;
  arReturn = DosCreateMutexSem(NULL, &(forks[i].hmtxSemaphore), ulAttributes, b32InitState);
  if (arReturn != 0)
  {
   fprintf(stderr, "\nCannot create a semaphore, return code = %d\n", arReturn);
   bError = TRUE;
  } /* if */
 } /* for i */

 arReturn = DosCreateMutexSem(NULL, &hmtxPrintMutex, ulAttributes, b32InitState);
 if (arReturn != 0)
 {
  fprintf(stderr, "\nCannot create a semaphore, return code = %d\n", arReturn);
  bError = TRUE;
 } /* if */

 arReturn = DosCreateMutexSem(NULL, &hmtxThreadCountMutex, ulAttributes, b32InitState);
 if (arReturn != 0)
 {
  fprintf(stderr, "\nCannot create a semaphore, return code = %d\n", arReturn);
  bError = TRUE;
 } /* if */

 arReturn = DosCreateEventSem(NULL, &hevBlockMainThreadSemaphore, ulAttributes, FALSE); /* FALSE == Not posted */
 if (arReturn != 0)
 {
  fprintf(stderr, "\nCannot create a semaphore, return code = %d\n", arReturn);
  bError = TRUE;
 } /* if */

 iNumberOfAlivePhilosophers = NO_OF_PHILOSOPHERS;

 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  philosophers[i].iID = i;
  philosophers[i].iState = STATE_DEAD;

  if (i == 0)
   philosophers[i].iRightFork = NO_OF_PHILOSOPHERS-1;
  else
   philosophers[i].iRightFork = i-1;
  philosophers[i].iLeftFork = i;

  if ((i % 2) == 0)
  {
   philosophers[i].iFirstFork = philosophers[i].iLeftFork;
   philosophers[i].iSecondFork = philosophers[i].iRightFork;
  }
  else
  {
   philosophers[i].iFirstFork = philosophers[i].iRightFork;
   philosophers[i].iSecondFork = philosophers[i].iLeftFork;
  }
 } /* for i */

 srand((unsigned int) time(NULL));

 return (!bError);
} /* init_data */



/*------------------------------------------------------------------------------------
destroy_data
------------------------------------------------------------------------------------*/
BOOL destroy_data(FORKS forks)
{
 INT            i;
 APIRET         arReturn;
 BOOL           bError = FALSE;

 printf("Destroying mutex semaphores... ");
 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  arReturn = DosCloseMutexSem(forks[i].hmtxSemaphore);
  if (arReturn != 0)
  {
   fprintf(stderr, "\nCannot close a semaphore, return code = %d\n", arReturn);
   bError = TRUE;
  } /* if */
 } /* for */

 arReturn = DosCloseMutexSem(hmtxPrintMutex);
 if (arReturn != 0)
 {
  fprintf(stderr, "\nCannot close a semaphore, return code = %d\n", arReturn);
  bError = TRUE;
 } /* if */

 arReturn = DosCloseMutexSem(hmtxThreadCountMutex);
 if (arReturn != 0)
 {
  fprintf(stderr, "\nCannot close a semaphore, return code = %d\n", arReturn);
  bError = TRUE;
 } /* if */

 arReturn = DosCloseEventSem(hevBlockMainThreadSemaphore);
 if (arReturn != 0)
 {
  fprintf(stderr, "\nCannot close a semaphore, return code = %d\n", arReturn);
  bError = TRUE;
 } /* if */

 return (!bError);
} /* destroy_data */



/*------------------------------------------------------------------------------------
fnPhilosopherThread
------------------------------------------------------------------------------------*/
VOID APIENTRY fnPhilosopherThread(ULONG ulThreadArgs)
{
 PHILOSOPHER_DATA       *ppdMyData;
 PHILOSOPHER            *me;
 INT                    i;
 APIRET                 arReturn1, arReturn2;

 ppdMyData = (PHILOSOPHER_DATA *) ulThreadArgs;
 me = &((*(ppdMyData->pPhilosophers))[ppdMyData->uiMyNumber]);

 for (i = 0; i < NO_OF_TURNS; i++)
 {
  /* Thinking... */
  me->iState = STATE_THINKING;
  print_forks(*(ppdMyData->pForks), *(ppdMyData->pPhilosophers));
  DosSleep(rand() % uiTimeToThink);

  /* Trying to get both forks... */
  me->iState = STATE_WAITING;
  print_forks(*(ppdMyData->pForks), *(ppdMyData->pPhilosophers));

  /* Get the first one */
  arReturn1 = DosRequestMutexSem(((*(ppdMyData->pForks))[me->iFirstFork]).hmtxSemaphore, SEM_INDEFINITE_WAIT);
  if (arReturn1 != 0)
  {
   fprintf(stderr, "\nCannot get a mutex semaphore, return code = %d\n", arReturn1);
   fprintf(stderr, "Thread terminated.");
   me->iState = STATE_DEAD;
   return;
  }
  ((*(ppdMyData->pForks))[me->iFirstFork]).iOwner = me->iID;
  print_forks(*(ppdMyData->pForks), *(ppdMyData->pPhilosophers));

  /* Get the second one */
  arReturn2 = DosRequestMutexSem(((*(ppdMyData->pForks))[me->iSecondFork]).hmtxSemaphore, SEM_INDEFINITE_WAIT);
  if (arReturn2 != 0)
  {
   fprintf(stderr, "\nCannot get a mutex semaphore, return code = %d\n", arReturn2);
   fprintf(stderr, "Thread terminated.");
   me->iState = STATE_DEAD;
   return;
  }
  ((*(ppdMyData->pForks))[me->iSecondFork]).iOwner = me->iID;
  print_forks(*(ppdMyData->pForks), *(ppdMyData->pPhilosophers));

  /* Eating... */
  me->iState = STATE_EATING;
  print_forks(*(ppdMyData->pForks), *(ppdMyData->pPhilosophers));
  DosSleep(rand() % uiTimeToEat);

  /* Releasing the forks... */
  me->iState = STATE_RELEASING;
  print_forks(*(ppdMyData->pForks), *(ppdMyData->pPhilosophers));

  /* Releasing the second fork */
  ((*(ppdMyData->pForks))[me->iSecondFork]).iOwner = NOBODY;
  arReturn2 = DosReleaseMutexSem(((*(ppdMyData->pForks))[me->iSecondFork]).hmtxSemaphore);
  if (arReturn2 != 0)
  {
   fprintf(stderr, "\nCannot release a mutex semaphore, return code = %d\n", arReturn2);
   fprintf(stderr, "Thread terminated.");
   me->iState = STATE_DEAD;
   return;
  }
  print_forks(*(ppdMyData->pForks), *(ppdMyData->pPhilosophers));

  /* Releasing the first fork */
  ((*(ppdMyData->pForks))[me->iFirstFork]).iOwner = NOBODY;
  arReturn1 = DosReleaseMutexSem(((*(ppdMyData->pForks))[me->iFirstFork]).hmtxSemaphore);
  if (arReturn1 != 0)
  {
   fprintf(stderr, "\nCannot release a mutex semaphore, return code = %d\n", arReturn1);
   fprintf(stderr, "Thread terminated.");
   me->iState = STATE_DEAD;
   return;
  }
  print_forks(*(ppdMyData->pForks), *(ppdMyData->pPhilosophers));
 } /* for i */

 me->iState = STATE_DEAD;
 print_forks(*(ppdMyData->pForks), *(ppdMyData->pPhilosophers));

 arReturn1 = DosRequestMutexSem(hmtxThreadCountMutex, SEM_INDEFINITE_WAIT);
 if (arReturn1 != 0)
 {
  fprintf(stderr, "\nCannot get a mutex semaphore, return code = %d\n", arReturn1);
  fprintf(stderr, "Thread terminated. Deadlock possible.");
  return;
 }
 iNumberOfAlivePhilosophers--;

 /* Restart the main thread? */
 if (iNumberOfAlivePhilosophers == 0)
 {
  arReturn2 = DosPostEventSem(hevBlockMainThreadSemaphore);
  if (arReturn2 != 0)
  {
   fprintf(stderr, "\nCannot post an event semaphore, return code = %d\n", arReturn2);
   fprintf(stderr, "Deadlock possible.");
  }
 }

 arReturn1 = DosReleaseMutexSem(hmtxThreadCountMutex);
 if (arReturn1 != 0)
  fprintf(stderr, "\nCannot release a mutex semaphore, return code = %d. Deadlock possible.\n", arReturn1);
 return;
} /* fnPhilosopherThread */



/*------------------------------------------------------------------------------------
create_threads
------------------------------------------------------------------------------------*/
BOOL create_threads(PHILOSOPHER_DATA philosopher_data[])
{
 BOOL           bError = FALSE;
 INT            i;
 TID            tidThreadIDs[NO_OF_PHILOSOPHERS];
 APIRET         arReturn;

 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  arReturn = DosCreateThread(&tidThreadIDs[i], fnPhilosopherThread, (ULONG) (&(philosopher_data[i])), THREAD_SUSPEND, 0x10000); /* 64 KB stack */
  if (arReturn != 0)
  {
   fprintf(stderr, "Cannot create a thread, rc = %d\n", arReturn);
   bError = TRUE;
  } /* if */
 } /* for */

 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  arReturn = DosResumeThread(tidThreadIDs[i]);
  if (arReturn != 0)
  {
   fprintf(stderr, "Cannot start a thread, rc = %d\n", arReturn);
   bError = TRUE;
  } /* if */
 } /* for */

 return (!bError);
} /* create_threads */



/*------------------------------------------------------------------------------------
hello
------------------------------------------------------------------------------------*/
VOID hello(VOID)
{
 printf("Dining Philosophers Example -- Uses native OS/2 threads and semaphores.\n");
 printf("Author: Jarda Kacer, jkacer@kiv.zcu.cz, Czech OS/2 User Group\n\n");
} /* hello */



/*------------------------------------------------------------------------------------
usage
------------------------------------------------------------------------------------*/
VOID usage(VOID)
{
 printf("Usage: %s MAX_THINKING_TIME MAX_EATING_TIME\n", pszProjectName);
 printf("MAX_THINKING_TIME ... Maximum time (in seconds) the philosophers will think.\n");
 printf("MAX_EATING_TIME ..... Maximum time (in seconds) the philosophers will eat.\n\n");
 printf("The actual time used for thinking/eating is a random number\n");
 printf("between 0 and MAX_?_TIME which is generated every time a philosopher\n");
 printf("attempts to think or eat.\n\n");
} /* usage */



/*------------------------------------------------------------------------------------
set_times
------------------------------------------------------------------------------------*/
BOOL set_times(UCHAR *argv[])
{
 BOOL   bError = FALSE;
 INT    i;

 for (i = 0; i < strlen(argv[1]); i++)
  if (isdigit(argv[1][i]) == 0)
   bError = TRUE;
 for (i = 0; i < strlen(argv[2]); i++)
  if (isdigit(argv[2][i]) == 0)
   bError = TRUE;

 if (bError == FALSE)
 {
  uiTimeToThink = atoi(argv[1]) * 1000;
  uiTimeToEat   = atoi(argv[2]) * 1000;

  printf("Using %u ms as maximum thinking time and %u ms as maximum eating time.\n", uiTimeToThink, uiTimeToEat);
  return TRUE;
 }
 else
 {
  printf("Invalid parameters specified.");
  return FALSE;
 }
} /* set_times */


/*------------------------------------------------------------------------------------
print_explanation
------------------------------------------------------------------------------------*/
VOID print_explanation(VOID)
{
 printf("\n");
 printf("%c ... Thinking\n", aucStates[STATE_THINKING]);
 printf("%c ... Waiting for a fork\n", aucStates[STATE_WAITING]);
 printf("%c ... Eating\n", aucStates[STATE_EATING]);
 printf("%c ... Releasing a fork\n", aucStates[STATE_RELEASING]);
 printf("%c ... Dead (without any fork)\n", aucStates[STATE_DEAD]);
 printf("\n");
 printf("| X | ... The philosopher has no fork.\n");
 printf(" /X | ... The philosopher has his right fork.\n");
 printf("| X\\  ... The philosopher has his left fork.\n");
 printf(" /X\\  ... The philosopher has both forks.\n");
 printf("\n");
} /* print_explanation */

/*------------------------------------------------------------------------------------
main
------------------------------------------------------------------------------------*/
INT main(INT argc, UCHAR *argv[])
{
 FORKS                  forks;
 PHILOSOPHERS           philosophers;
 PHILOSOPHER_DATA       philosopher_data[NO_OF_PHILOSOPHERS];

 BOOL                   bSuccess;
 APIRET                 arReturn;
 INT                    i;


 hello();
 if (argc != 3)
  usage();
 else
 {
  bSuccess = set_times(argv);

  if (bSuccess == TRUE)
  {
   /* Let's create all semaphores now */
   bSuccess = init_data(forks, philosophers);
   if (bSuccess == TRUE)
   {
    printf("OK\n");

    print_explanation();
    /* Every philosopher must get a unique number, the rest of data is shared */
    for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
    {
     philosopher_data[i].pForks = &forks;
     philosopher_data[i].pPhilosophers = &philosophers;
     philosopher_data[i].uiMyNumber = i;
    } /* for i */

    /* And let's start all philosophers */
    if (create_threads(philosopher_data) == TRUE)
    {
     /* The main thread must wait for the philosophers to finish */
     arReturn = DosWaitEventSem(hevBlockMainThreadSemaphore, SEM_INDEFINITE_WAIT);
     if (arReturn != 0)
     {
      fprintf(stderr, "\nCannot wait for an event semaphore, return code = %d\n", arReturn);
      fprintf(stderr, "Exiting.");
      return (2);
     } /* if arReturn */
    } /* if create_threads */
    else
     return (1);
   } /* init_data OK */
   else
    printf("An error occured during program startup.\n");

   /* Let's destroy all semaphores now */
   if ((bSuccess = destroy_data(forks)) == TRUE)
    printf("OK\n");
   else
    printf("An error occured during program shutdown.\n");
  } /* if set_times OK */
 } /* else argc */

 return (0);
} /* main */

