/*
Dining philosophers example. The program uses POSIX threads and semaphores.
You must install the POSIX library (pthreads004.zip), available at Hobbes, to your EMX package.
Output on console -- text mode application. The program terminates when all philosophers have
already eaten and thought 5 times. You can redefine the NO_OF_TURNS constant or rewrite
the philosopher_thread function (replace for(...) with while(true)) to change this behavior.

Author: Jarda Kacer, Czech OS/2 User Group
        mailto:jkacer@kiv.zcu.cz
        http://home.zcu.cz/~jkacer
*/


/*------------------------------------------------------------------------------------*/
/* Defined constants */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <pthread.h>


#define NO_OF_PHILOSOPHERS      5
#define NOBODY                  -1

#define NO_OF_STATES            5

#define STATE_THINKING          0
#define STATE_WAITING           1
#define STATE_EATING            2
#define STATE_RELEASING         3
#define STATE_DEAD              4

#define NO_OF_TURNS             5



/*------------------------------------------------------------------------------------*/
/* Types */


typedef struct _FORK
        {
         int			owner;
         pthread_mutex_t	semaphore;
        }
        FORK;

typedef FORK FORKS[NO_OF_PHILOSOPHERS];

typedef struct _PHILOSOPHER
        {
         int    id;
         int    state;

         int    left_fork;
         int    right_fork;

         int    first_fork;
         int    second_fork;
        }
        PHILOSOPHER;

typedef PHILOSOPHER PHILOSOPHERS[NO_OF_PHILOSOPHERS];

typedef struct _PHILOSOPHER_DATA
        {
         FORKS          *pointer_forks;
         PHILOSOPHERS   *pointer_philosophers;
         unsigned int   my_number;
        }
        PHILOSOPHER_DATA;



/*------------------------------------------------------------------------------------*/
/* Global data */


pthread_mutex_t		print_mutex;
pthread_mutex_t		thread_count_mutex;
int     		number_of_alive_philosophers;
pthread_cond_t		block_main_thread_condition;
char			states[NO_OF_STATES] = {'T', 'W', 'E', 'R', 'D'};
unsigned int    	time_to_think;
unsigned int    	time_to_eat;
char    		*project_name = "dining_philosophers_c_posix.exe";



/*------------------------------------------------------------------------------------*/
/* Functions */


/*------------------------------------------------------------------------------------
print_1_philosopher
------------------------------------------------------------------------------------*/
void print_1_philosopher(FORKS forks, PHILOSOPHER philosopher)
{
 /* Right fork */
 if (forks[philosopher.right_fork].owner == philosopher.id)
  printf(" /");
 else
  printf("  ");

 printf("%c", states[philosopher.state]);

 /* Left fork */
 if (forks[philosopher.left_fork].owner == philosopher.id)
  printf("\\ ");
 else
  printf("  ");
} /* print_1_philosopher */



/*------------------------------------------------------------------------------------
print_forks
------------------------------------------------------------------------------------*/
void print_forks(FORKS forks, PHILOSOPHERS philosophers)
{
 int            i;
 int		retval;

 if ((retval = pthread_mutex_lock(&print_mutex)) != 0)
 {
  fprintf(stderr, "\nCannot get a mutex semaphore, return code = %d\n", retval);
  return;
 }
 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  print_1_philosopher(forks, philosophers[i]);
  if (forks[i].owner == NOBODY)
   printf("|");
  else
   printf(" ");
 }
 printf("\n");
 if ((retval = pthread_mutex_unlock(&print_mutex)) != 0)
 {
  fprintf(stderr, "\nCannot release a mutex semaphore, return code = %d\n", retval);
  return;
 }
} /* print_forks */



/*------------------------------------------------------------------------------------
init_data
------------------------------------------------------------------------------------*/
int init_data(FORKS forks, PHILOSOPHERS philosophers)
{
 int            i;
 int		retval;
 int		error = 0;

 printf("Creating semaphores... ");

 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  forks[i].owner = NOBODY;
  retval = pthread_mutex_init(&(forks[i].semaphore), NULL);
  if (retval != 0)
  {
   fprintf(stderr, "\nCannot create a semaphore, return code = %d\n", retval);
   error = 1;
  } /* if */
 } /* for i */

 retval = pthread_mutex_init(&print_mutex, NULL);
 if (retval != 0)
 {
  fprintf(stderr, "\nCannot create a semaphore, return code = %d\n", retval);
  error = 1;
 } /* if */

 retval = pthread_mutex_init(&thread_count_mutex, NULL);
 if (retval != 0)
 {
  fprintf(stderr, "\nCannot create a semaphore, return code = %d\n", retval);
  error = 1;
 } /* if */

 retval = pthread_cond_init(&block_main_thread_condition, NULL);
 if (retval != 0)
 {
  fprintf(stderr, "\nCannot create a condition variable, return code = %d\n", retval);
  error = 1;
 } /* if */

 number_of_alive_philosophers = NO_OF_PHILOSOPHERS;

 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  philosophers[i].id = i;
  philosophers[i].state = STATE_DEAD;

  if (i == 0)
   philosophers[i].right_fork = NO_OF_PHILOSOPHERS-1;
  else
   philosophers[i].right_fork = i-1;
  philosophers[i].left_fork = i;

  if ((i % 2) == 0)
  {
   philosophers[i].first_fork = philosophers[i].left_fork;
   philosophers[i].second_fork = philosophers[i].right_fork;
  }
  else
  {
   philosophers[i].first_fork = philosophers[i].right_fork;
   philosophers[i].second_fork = philosophers[i].left_fork;
  }
 } /* for i */

 srand((unsigned int) time(NULL));

 return (1-error);
} /* init_data */



/*------------------------------------------------------------------------------------
destroy_data
------------------------------------------------------------------------------------*/
int destroy_data(FORKS forks)
{
 int            i;
 int		retval;
 int		error = 0;

 printf("Destroying mutex semaphores... ");
 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  retval = pthread_mutex_destroy(&forks[i].semaphore);
  if (retval != 0)
  {
   fprintf(stderr, "\nCannot close a semaphore, return code = %d\n", retval);
   error = 1;
  } /* if */
 } /* for */

 retval = pthread_mutex_destroy(&print_mutex);
 if (retval != 0)
 {
  fprintf(stderr, "\nCannot close a semaphore, return code = %d\n", retval);
  error = 1;
 } /* if */

 retval = pthread_mutex_destroy(&thread_count_mutex);
 if (retval != 0)
 {
  fprintf(stderr, "\nCannot close a semaphore, return code = %d\n", retval);
  error = 1;
 } /* if */

 retval = pthread_cond_destroy(&block_main_thread_condition);
 if (retval != 0)
 {
  fprintf(stderr, "\nCannot close a condition variable, return code = %d\n", retval);
  error = 1;
 } /* if */

 return (1-error);
} /* destroy_data */



/*------------------------------------------------------------------------------------
philosopher_thread
------------------------------------------------------------------------------------*/
void *philosopher_thread(void *data)
{
 PHILOSOPHER_DATA       *my_data;
 PHILOSOPHER            *me;
 int                    i;
 int                    retval1, retval2;

 my_data = (PHILOSOPHER_DATA *) data;
 me = &((*(my_data->pointer_philosophers))[my_data->my_number]);

 for (i = 0; i < NO_OF_TURNS; i++)
 {
  /* Thinking... */
  me->state = STATE_THINKING;
  print_forks(*(my_data->pointer_forks), *(my_data->pointer_philosophers));
  /*
   *  _sleep2() is an EMX-specific replacement for ANSI sleep().
   *  _sleep2() is more convenient for our purposes since it can suspend a thread for milliseconds instead
   *  of seconds as sleep() does it. If we used sleep() and "1 1" was passed as command-line attributes
   *  the threads would never wait here: rand() % 1 == 0
   */
  _sleep2(rand() % time_to_think);

  /* Trying to get both forks... */
  me->state = STATE_WAITING;
  print_forks(*(my_data->pointer_forks), *(my_data->pointer_philosophers));

  /* Get the first one */
  retval1 = pthread_mutex_lock(&((*(my_data->pointer_forks))[me->first_fork]).semaphore);
  if (retval1 != 0)
  {
   fprintf(stderr, "\nCannot get a mutex semaphore, return code = %d\n", retval1);
   fprintf(stderr, "Thread terminated.");
   me->state = STATE_DEAD;
   return (NULL);
  }
  ((*(my_data->pointer_forks))[me->first_fork]).owner = me->id;
  print_forks(*(my_data->pointer_forks), *(my_data->pointer_philosophers));

  /* Get the second one */
  retval2 = pthread_mutex_lock(&((*(my_data->pointer_forks))[me->second_fork]).semaphore);
  if (retval2 != 0)
  {
   fprintf(stderr, "\nCannot get a mutex semaphore, return code = %d\n", retval2);
   fprintf(stderr, "Thread terminated.");
   me->state = STATE_DEAD;
   return (NULL);
  }
  ((*(my_data->pointer_forks))[me->second_fork]).owner = me->id;
  print_forks(*(my_data->pointer_forks), *(my_data->pointer_philosophers));

  /* Eating... */
  me->state = STATE_EATING;
  print_forks(*(my_data->pointer_forks), *(my_data->pointer_philosophers));
  /*
   *  _sleep2() is an EMX-specific replacement for ANSI sleep().
   *  _sleep2() is more convenient for our purposes since it can suspend a thread for milliseconds instead
   *  of seconds as sleep() does it. If we used sleep() and "1 1" was passed as command-line attributes
   *  the threads would never wait here: rand() % 1 == 0
   */
  _sleep2(rand() % time_to_eat);

  /* Releasing the forks... */
  me->state = STATE_RELEASING;
  print_forks(*(my_data->pointer_forks), *(my_data->pointer_philosophers));

  /* Releasing the second fork */
  ((*(my_data->pointer_forks))[me->second_fork]).owner = NOBODY;
  retval2 = pthread_mutex_unlock(&((*(my_data->pointer_forks))[me->second_fork]).semaphore);
  if (retval2 != 0)
  {
   fprintf(stderr, "\nCannot release a mutex semaphore, return code = %d\n", retval2);
   fprintf(stderr, "Thread terminated.");
   me->state = STATE_DEAD;
   return (NULL);
  }
  print_forks(*(my_data->pointer_forks), *(my_data->pointer_philosophers));

  /* Releasing the first fork */
  ((*(my_data->pointer_forks))[me->first_fork]).owner = NOBODY;
  retval1 = pthread_mutex_unlock(&((*(my_data->pointer_forks))[me->first_fork]).semaphore);
  if (retval1 != 0)
  {
   fprintf(stderr, "\nCannot release a mutex semaphore, return code = %d\n", retval1);
   fprintf(stderr, "Thread terminated.");
   me->state = STATE_DEAD;
   return (NULL);
  }
  print_forks(*(my_data->pointer_forks), *(my_data->pointer_philosophers));
 } /* for i */

 me->state = STATE_DEAD;
 print_forks(*(my_data->pointer_forks), *(my_data->pointer_philosophers));

 retval1 = pthread_mutex_lock(&thread_count_mutex);
 if (retval1 != 0)
 {
  fprintf(stderr, "\nCannot get a mutex semaphore, return code = %d\n", retval1);
  fprintf(stderr, "Thread terminated. Deadlock possible.");
  return (NULL);
 }
 number_of_alive_philosophers--;

 /* Restart the main thread? */
 if (number_of_alive_philosophers == 0)
 {
  retval2 = pthread_cond_signal(&block_main_thread_condition);
  if (retval2 != 0)
  {
   fprintf(stderr, "\nCannot post an event semaphore, return code = %d\n", retval2);
   fprintf(stderr, "Deadlock possible.");
  }
 }

 retval1 = pthread_mutex_unlock(&thread_count_mutex);
 if (retval1 != 0)
  fprintf(stderr, "\nCannot release a mutex semaphore, return code = %d. Deadlock possible.\n", retval1);
 return (NULL);
} /* philosopher_thread */



/*------------------------------------------------------------------------------------
create_threads
------------------------------------------------------------------------------------*/
int create_threads(PHILOSOPHER_DATA philosopher_data[])
{
 int            error = 0;
 int            i;
 int		retval;
 pthread_t	threads[NO_OF_PHILOSOPHERS];

 for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
 {
  retval = pthread_create(&threads[i], NULL, philosopher_thread, (void *) &philosopher_data[i]);
  if (retval != 0)
  {
   fprintf(stderr, "Cannot create a thread, rc = %d\n", retval);
   error = 1;
  } /* if */
 } /* for */

 return (1-error);
} /* create_threads */



/*------------------------------------------------------------------------------------
hello
------------------------------------------------------------------------------------*/
void hello(void)
{
 printf("Dining Philosophers Example -- Uses POSIX threads and semaphores.\n");
 printf("Author: Jarda Kacer, jkacer@kiv.zcu.cz, Czech OS/2 User Group\n\n");
} /* hello */



/*------------------------------------------------------------------------------------
usage
------------------------------------------------------------------------------------*/
void usage(void)
{
 printf("Usage: %s MAX_THINKING_TIME MAX_EATING_TIME\n", project_name);
 printf("MAX_THINKING_TIME ... Maximum time (in seconds) the philosophers will think.\n");
 printf("MAX_EATING_TIME ..... Maximum time (in seconds) the philosophers will eat.\n\n");
 printf("The actual time used for thinking/eating is a random number\n");
 printf("between 0 and MAX_?_TIME which is generated every time a philosopher\n");
 printf("attempts to think or eat.\n\n");
} /* usage */



/*------------------------------------------------------------------------------------
set_times
------------------------------------------------------------------------------------*/
int set_times(char *argv[])
{
 int    error = 0;
 int    i;

 for (i = 0; i < strlen(argv[1]); i++)
  if (isdigit((int) argv[1][i]) == 0)
   error = 1;
 for (i = 0; i < strlen(argv[2]); i++)
  if (isdigit((int) argv[2][i]) == 0)
   error = 1;

 if (error == 0)
 {
  time_to_think = atoi(argv[1]) * 1000;
  time_to_eat   = atoi(argv[2]) * 1000;

  printf("Using %u ms as maximum thinking time and %u ms as maximum eating time.\n", time_to_think, time_to_eat);
  return 1;
 }
 else
 {
  printf("Invalid parameters specified.");
  return 0;
 }
} /* set_times */


/*------------------------------------------------------------------------------------
print_explanation
------------------------------------------------------------------------------------*/
void print_explanation(void)
{
 printf("\n");
 printf("%c ... Thinking\n", states[STATE_THINKING]);
 printf("%c ... Waiting for a fork\n", states[STATE_WAITING]);
 printf("%c ... Eating\n", states[STATE_EATING]);
 printf("%c ... Releasing a fork\n", states[STATE_RELEASING]);
 printf("%c ... Dead (without any fork)\n", states[STATE_DEAD]);
 printf("\n");
 printf("| X | ... The philosopher has no fork.\n");
 printf(" /X | ... The philosopher has his right fork.\n");
 printf("| X\\  ... The philosopher has his left fork.\n");
 printf(" /X\\  ... The philosopher has both forks.\n");
 printf("\n");
} /* print_explanation */

/*------------------------------------------------------------------------------------
main
------------------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
 FORKS                  forks;
 PHILOSOPHERS           philosophers;
 PHILOSOPHER_DATA       philosopher_data[NO_OF_PHILOSOPHERS];

 int                    success;
 int                    retval;
 int                    i;


 hello();
 if (argc != 3)
  usage();
 else
 {
  success = set_times(argv);

  if (success == 1)
  {
   /* Let's create all semaphores now */
   success = init_data(forks, philosophers);
   if (success == 1)
   {
    printf("OK\n");

    print_explanation();
    /* Every philosopher must get a unique number, the rest of data is shared */
    for (i = 0; i < NO_OF_PHILOSOPHERS; i++)
    {
     philosopher_data[i].pointer_forks = &forks;
     philosopher_data[i].pointer_philosophers = &philosophers;
     philosopher_data[i].my_number = i;
    } /* for i */

    /* And let's start all philosophers */
    if (create_threads(philosopher_data) == 1)
    {
     /* The main thread must wait for the philosophers to finish */
     if ((retval = pthread_mutex_lock(&thread_count_mutex)) != 0)
     {
      fprintf(stderr, "\nCannot get a mutex semaphore, return code = %d\n", retval);
      fprintf(stderr, "Exiting.");
      return (2);
     }

     retval = pthread_cond_wait(&block_main_thread_condition, &thread_count_mutex);
     if (retval != 0)
     {
      fprintf(stderr, "\nCannot wait for an event semaphore, return code = %d\n", retval);
      fprintf(stderr, "Exiting.");
      return (2);
     } /* if retval */

     if ((retval = pthread_mutex_unlock(&thread_count_mutex)) != 0)
     {
      fprintf(stderr, "\nCannot release a mutex semaphore, return code = %d\n", retval);
      fprintf(stderr, "Exiting.");
      return (2);
     }
    } /* if create_threads */
    else
     return (1);
   } /* init_data OK */
   else
    printf("An error occured during program startup.\n");

   /* Let's destroy all semaphores now */
   if ((success = destroy_data(forks)) == 1)
    printf("OK\n");
   else
    printf("An error occured during program shutdown.\n");
  } /* if set_times OK */
 } /* else argc */

 return (0);
} /* main */

